# coding: utf-8
"""A SRU client.

:copyright: Copyright 2020 Andreas Lüschow
"""
import inspect
import logging
import time

import requests

from srupy.iterator import BaseSRUIterator, SRUResponseIterator
from srupy.response import SRUResponse
from .models import Explain

logger = logging.getLogger(__name__)

SRU_NAMESPACE = '{http://docs.oasis-open.org/ns/search-ws/sruResponse}'


class SRUpy(object):
    """Client for sending SRU requests.

    Use it like this::
        >>> py_sru = SRUpy('http://elis.da.ulcc.ac.uk/cgi/oai2')
        >>> records = py_sru.get_records()
        >>> records.next()
    :param endpoint: The endpoint of the SRU server.
    :type endpoint: str
    :param http_method: Method used for requests (GET or POST, default: GET).
    :type http_method: str
    :param protocol_version: The SRU protocol version.
    :type protocol_version: str
    :param iterator: The type of the returned iterator
           (default: :class:`srupy.iterator.SRUResponseIterator`)
    :param max_retries: Number of retry attempts if an HTTP
                        request fails (default: 0 = request
                        only once). SRUpy will use the value
                        from the retry-after header (if present)
                        and will wait the specified number of
                        seconds between retries.
    :type max_retries: int
    :param retry_status_codes:  HTTP status codes to
                                retry (default will only retry on 503)
    :type retry_status_codes: iterable
    :param default_retry_after: default number of seconds to wait
                                between retries in case no retry-after
                                header is found on the response
                                (defaults to 60 seconds)
    :type default_retry_after: int
    :param encoding:    Can be used to override the encoding used
                        when decoding the server response. If not
                        specified, `requests` will use the encoding
                        returned by the server in the `content-type`
                        header. However, if the `charset`
                        information is missing, `requests` will fallback to
                        `'ISO-8859-1'`.
    :type encoding:      str
    :param request_args: Arguments to be passed to requests when issuing HTTP
                         requests. See the `documentation of requests
                         <http://docs.python-requests.org/en/master/api/#main-interface>`_
                         for all available parameters.
    """

    def __init__(self, endpoint,
                 http_method='GET',
                 protocol_version='2.0',
                 iterator=SRUResponseIterator,
                 max_retries=0,
                 retry_status_codes=None,
                 default_retry_after=60,
                 encoding=None,
                 **request_args):
        """Docstring."""
        self.endpoint = endpoint
        if http_method not in ['GET', 'POST']:
            raise ValueError("Invalid HTTP method: %s! Must be GET or POST.")
        if protocol_version not in ['2.0', '1.2', '1.1']:
            raise ValueError(
                "Invalid protocol version: %s! Must be 2.0, 1.2 or 1.1")
        self.http_method = http_method
        self.protocol_version = protocol_version
        if inspect.isclass(iterator) and issubclass(iterator, BaseSRUIterator):
            self.iterator = iterator
        else:
            raise TypeError(
                "Argument 'iterator' must be subclass of %s"
                % BaseSRUIterator.__name__)
        self.max_retries = max_retries
        self.retry_status_codes = retry_status_codes or [503]
        self.default_retry_after = default_retry_after
        self.sru_namespace = SRU_NAMESPACE
        self.encoding = encoding
        self.request_args = request_args

    def harvest(self, **kwargs):
        """Make HTTP requests to the SRU server.

        :param kwargs: SRU HTTP parameters.
        :rtype: :class:`srupy.SRUResponse`
        """
        http_response = self._request(kwargs)
        for _ in range(self.max_retries):
            if self._is_error_code(http_response.status_code) \
                    and http_response.status_code in self.retry_status_codes:
                retry_after = self.get_retry_after(http_response)
                logger.warning(
                    "HTTP %d! Retrying after %d seconds..."
                    % (http_response.status_code, retry_after))
                time.sleep(retry_after)
                http_response = self._request(kwargs)
        http_response.raise_for_status()
        if self.encoding:
            http_response.encoding = self.encoding
        return SRUResponse(http_response, params=kwargs)

    def _request(self, kwargs):
        """Docstring."""
        if self.http_method == 'GET':
            return requests.get(self.endpoint,
                                params=kwargs, **self.request_args)
        return requests.post(self.endpoint,
                             data=kwargs, **self.request_args)

    def get_records(self, **kwargs):
        """Issue a SRU request to fetch records.

        :rtype: :class:`srupy.iterator.BaseSRUIterator`
        """
        # TODO: Default Parameter hier eintragen
        # Default Wert für z. B. recordSchema wird vom
        # jeweiligen Server festgelegt;
        # also vorher ExplainFile anschauen!
        params = {
                # 'query': 'dog and cat and mouse',
                # 'queryType': 'cql',
                # default value is 1
                'startRecord': 1,

                # default value is determined by the server
                # for the sake of multi page browsing,
                # we use 10 as default, however
                # 'maximumRecords': 10,
                # 'recordSchema': 'mods',

                # 'record_XML_escaping' = True
                # resultSetTTL = True
                # Stylesheet = True
                # extension_parameters
                # sortKeys = True
                # facet_parameters
                # renderedBy = True
                # httpAccept = True
                # responseType = True
                # recordPacking = True
            }
        params.update(kwargs)

        if 'query' not in params.keys():
            raise KeyError("Request parameter 'query' must be set")

        return self.iterator(self, params)

    def explain(self):
        """Issue an Explain request to a SRU server.

        :rtype: :class:`srupy.models.Explain`
        """
        return Explain(self.harvest())

    def get_retry_after(self, http_response):
        """Docstring."""
        if http_response.status_code == 503:
            try:
                return int(http_response.headers.get('retry-after'))
            except TypeError:
                return self.default_retry_after
        return self.default_retry_after

    @staticmethod
    def _is_error_code(status_code):
        """Docstring."""
        return status_code >= 400



# coding: utf-8
"""Collects classes for SRU-specific entities.

:copyright: Copyright 2020 Andreas Lüschow
"""

from lxml import etree
from .util import get_namespace, etree_to_dict, etree_to_dict_without_ns


class SRURecord(object):
    """A generic SRU record.

    :param xml: XML representation of the record.
    :param strip_ns: Flag for whether to remove the namespaces from the
                     element names in the dictionary representation.
    """

    def __init__(self, xml, strip_ns=True):
        """Docstring."""
        super(SRURecord, self).__init__()

        #: The original parsed XML
        self.xml = xml
        self._strip_ns = strip_ns
        self._sru_namespace = get_namespace(self.xml)

    def __bytes__(self):
        """Docstring."""
        return etree.tounicode(self.xml).encode("utf8")

    def __str__(self):
        """Docstring."""
        return self.__unicode__()

    def __unicode__(self):
        """Docstring."""
        return etree.tounicode(self.xml)

    @property
    def raw(self):
        """Return the original XML as unicode."""
        return etree.tounicode(self.xml)


class Explain(SRURecord):
    """Represent a SRU Explain record.

    This object differs from the other entities in that is has to be created
    from a :class:`srupy.response.SRUResponse` instead of an XML element.
    :param explain_response: The response for an Explain request.
    :type explain_response: :class:`srupy.SRUResponse`
    """

    def __init__(self, explain_response):
        """Docstring."""
        super(Explain, self).__init__(explain_response.xml, strip_ns=True)
        self._explain_dict = self.get_explain_data()
        # TODO: Ist echoedExplainRequest vorgeschrieben bzw.
        #  im Standard vorgesehen?
        self.echo = EchoedRequest(self.xml.find(
            './/' + self._sru_namespace + 'echoedExplainRequest'))

    def __iter__(self):
        """Docstring."""
        return iter(self._explain_dict.items())

    def get_explain_data(self):
        """Docstring."""
        return etree_to_dict_without_ns(
            self.xml.find(
                './/' + self._sru_namespace + 'recordData'
            ).getchildren()[0]) \
            if self._strip_ns \
            else etree_to_dict(
            self.xml.find(
                './/' + self._sru_namespace + 'recordData'
            ).getchildren()[0])


class EchoedRequest(SRURecord):
    """Represent an SRU Echoed Request element.

    :param sru_xml: The SRU XML response.
    :type sru_xml: :class:`lxml.etree._Element`
    """

    def __init__(self, sru_xml):
        """Docstring."""
        super(EchoedRequest, self).__init__(sru_xml, strip_ns=True)

        _version = self.xml.find(self._sru_namespace + 'version')
        _query = self.xml.find(self._sru_namespace + 'query')
        _start_record = self.xml.find(self._sru_namespace + 'startRecord')
        _maximum_records = self.xml.find(
            self._sru_namespace + 'maximumRecords'
        )
        _record_xml_escaping = self.xml.find(
            self._sru_namespace + 'recordXMLEscaping'
        )
        _record_schema = self.xml.find(self._sru_namespace + 'recordSchema')
        _base_url = self.xml.find(self._sru_namespace + 'baseUrl')
        _xquery = self.xml.find(self._sru_namespace + 'xQuery')

        self.version = getattr(_version, 'text', None)
        self.query = getattr(_query, 'text', None)
        self.startRecord = getattr(_start_record, 'text', None)
        self.maximumRecords = getattr(_maximum_records, 'text', None)
        self.recordXMLEscaping = getattr(_record_xml_escaping, 'text', None)
        self.recordSchema = getattr(_record_schema, 'text', None)
        self.baseUrl = getattr(_base_url, 'text', None)
        self.xQuery = getattr(_xquery, 'text', None)

        self._echo_dict = {
            "version": self.version,
            "query": self.query,
            "startRecord": self.startRecord,
            "maximumRecords": self.maximumRecords,
            "recordXMLEscaping": self.recordXMLEscaping,
            "recordSchema": self.recordSchema,
            "baseUrl": self.baseUrl,
            "xQuery": self.xQuery
        }

    def __iter__(self):
        """Docstring."""
        return iter(self._echo_dict.items())


class Record(SRURecord):
    """Represent a SRU record.

    :param record_element: The XML element 'record'.
    :type record_element: :class:`lxml.etree._Element`
    :param strip_ns: Flag for whether to remove the namespaces from the
                     element names.
    """

    def __init__(self, record_element, strip_ns=True):
        """Docstring."""
        super(Record, self).__init__(record_element, strip_ns=strip_ns)
        self.record_data = self.get_record_data()

    def __iter__(self):
        """Docstring."""
        return iter(self.record_data.items())

    def get_record_data(self):
        """Docstring."""
        if self._strip_ns:
            return etree_to_dict_without_ns(
                self.xml.find(
                    './/' + self._sru_namespace + 'recordData'
                ).getchildren()[0])
        else:
            return etree_to_dict(
                self.xml.find(
                    './/' + self._sru_namespace + 'recordData'
                ).getchildren()[0])



# coding: utf-8
"""Collects classes for iterating over SRU responses.

:copyright: Copyright 2020 Andreas Lüschow
"""

from srupy import sruexceptions
from .models import (Record, EchoedRequest)


class BaseSRUIterator(object):
    """Iterator over SRU records.

    :param srupy: The SRUpy object that issued the first request.
    :type srupy: :class:`srupy.app.SRUpy`
    :param params: The SRU arguments.
    :type params:  dict
    """

    def __init__(self, srupy, params):
        """Docstring."""
        self.srupy = srupy
        self.params = params
        self._next_response()

    def __iter__(self):
        """Docstring."""
        return self

    def __next__(self):
        """Docstring."""
        return self.next()

    def _next_response(self):
        """Get the next response from the SRU server."""
        params = self.params
        self.sru_response = self.srupy.harvest(**params)
        # TODO
        error = self.sru_response.xml.find(
            './/' + self.srupy.sru_namespace + 'error')
        if error is not None:
            code = error.attrib.get('code', 'UNKNOWN')
            description = error.text or ''
            try:
                raise getattr(
                    sruexceptions, code[0].upper() + code[1:])(description)
            except AttributeError:
                raise sruexceptions.SRUError(description)

    def next(self):
        """Must be implemented by subclasses."""
        raise NotImplementedError


class SRUResponseIterator(BaseSRUIterator):
    """Handle SRU responses."""

    def __init__(self, srupy, params):
        """Docstring."""

        # handle large values for maximumRecords parameter
        self._multipage_threshold = 100
        self._total_records_wanted = int(params.get("maximumRecords"))
        if self._total_records_wanted > self._multipage_threshold:
            params.update({
                'maximumRecords': self._multipage_threshold,
            })

        super(SRUResponseIterator, self).__init__(srupy, params)

        # get basic information
        try:
            self.number_of_records = int(self.sru_response.xml.find(
                './/' + self.srupy.sru_namespace + 'numberOfRecords').text)
        except:
            self.number_of_records = 0

        self.echo = EchoedRequest(self.sru_response.xml.find(
            './/' + self.srupy.sru_namespace + 'echoedSearchRetrieveRequest'))

        # handle Request parameters
        self._maximum_records = int(params.get('maximumRecords'))
        self._start_record = int(params.get('startRecord'))

    def _next_response(self):
        """Docstring."""
        super(SRUResponseIterator, self)._next_response()
        self._records = self.sru_response.xml.iterfind(
            './/' + self.srupy.sru_namespace + 'record')

    def next(self):
        """Return the next record."""
        while True:
            for r in self._records:
                return Record(r)
            self._handle_multiple_pages()
            self._next_response()

    def _handle_multiple_pages(self):
        """Get startRecord for next Request.

        (if not all available records were fetched)
        """
        if self._start_record + self._maximum_records < self._total_records_wanted:
            self._start_record += self._maximum_records
            self.params.update({
                'startRecord': self._start_record,
            })
        else:
            raise StopIteration
